#include "stdafx.h"

#include "CubeLut.h"
#include "Image.h"


// generate a "no transform lut"
void CubeLut::GenerateNoTransform() {
	m_dwSize = 2;
	m_cube = unique_ptr< RGB< double > >( new RGB< double >[ m_dwSize * m_dwSize * m_dwSize ] );
	for( dword ib = 0; ib < m_dwSize; ib++ )
		for( dword ig = 0; ig < m_dwSize; ig++ )
			for( dword ir = 0; ir < m_dwSize; ir++ )
				_Set( ir, ig, ib, RGB< double >( double( ir ) / double( m_dwSize - 1 ), double( ig ) / double( m_dwSize - 1 ), double( ib ) / double( m_dwSize - 1 ) ) );
}


// load a cube file
const bool CubeLut::Load( const wstring & _sFileName ) {
	// load file content
	ifstream srcFile;
	srcFile.open( _sFileName.c_str(), ios::in | ios::binary );
	if( !srcFile ) {
		wcout << L"[ERROR] fail to open cube file: '" << _sFileName << L"'\n";
		return false;
	}
	srcFile.seekg( 0, ios::end );
	string srcData;
	srcData.resize( ( unsigned int ) srcFile.tellg() );
	srcFile.seekg( 0, ios::beg );
	srcFile.read( &srcData[ 0 ], srcData.size() );
	srcFile.close();
	
	// extract LUT size
	string sSize( "LUT_3D_SIZE" );
	size_t szPos = srcData.find( sSize );
	szPos += sSize.length();
	m_dwSize = _GetInt( srcData, szPos );

	// allocate the cube
	m_cube = unique_ptr< RGB< double > >( new RGB< double >[ m_dwSize * m_dwSize * m_dwSize ] );

	// skip the junk stuff
	_SkipTheJunk( srcData, szPos );

	// read all elements of the lookup table
	for( dword ib = 0; ib < m_dwSize; ib++ ) {
		for( dword ig = 0; ig < m_dwSize; ig++ ) {
			for( dword ir = 0; ir < m_dwSize; ir++ ) {
				double r = _GetVal( srcData, szPos );
				double g = _GetVal( srcData, szPos );
				double b = _GetVal( srcData, szPos );
				_Set( ir, ig, ib, RGB< double >( r, g, b ) );
			}
		}
	}
	return true;
}


// get int
dword CubeLut::_GetInt( const string & _srcData, size_t & _szPos ) {
	char c = 0;
	while( c < '0' || c > '9' ) c = _srcData.at( _szPos++ );
	string sVal;
	while( c >= '0' && c <= '9' ) { sVal += c; c = _srcData.at( _szPos++ ); }
	return atoi( sVal.c_str() );
}


// get double
double CubeLut::_GetVal( const string & _srcData, size_t & _szPos ) {
	char c = 0;
	while( c < '0' || c > '9' ) c = _srcData.at( _szPos++ );
	string sVal;
	while( ( c >= '0' && c <= '9' ) || c == '.' ) { sVal += c; c = _srcData.at( _szPos++ ); }
	return atof( sVal.c_str() );
}


// skip the junk stuff
void CubeLut::_SkipTheJunk( const string & _srcData, size_t & _szPos ) {
	char c = 0;
	string sInfo;
	bool bInfo = false;
	bool bIgnore = false;

	// skip or ignore chars
	while( c < '0' || c > '9' || bIgnore ) {
		c = _srcData.at( _szPos++ );

		// start info detection
		if( !bInfo && ( c == 'D' || c == 'T' ) ) {
			bInfo = true;
			sInfo.clear();
		}

		// ignore the line if it's a DOMAIN or TITLE
		if( bInfo ) {
			sInfo += c;
			if( sInfo == "DOMAIN_" || sInfo == "TITLE" )
				bIgnore = true;
		}

		// ignore comment lines
		if( c == '#' )
			bIgnore = true;

		// reset ignoring at the end of the line
		if( bIgnore && c == '\n' ) {
			bIgnore = false;
			bInfo = false;
		}
	}
	_szPos--;
}


// apply a lookup table on image
void CubeLut::Apply( const Image & _img ) {
	Assert( m_dwSize != 0 );
	Assert( _img.GetWidth() != 0 && _img.GetHeight() != 0 );

	// get image data access
	byte * pData = _img;

	// for each components
	for( dword y = 0; y < _img.GetHeight(); y++ ) {
		dword dwYS = y * _img.GetWidth();
		for( dword x = 0; x < _img.GetWidth(); x++ ) {
			dword dwPos = ( dwYS + x ) * 3;
			// get RGB components
			RGB< byte > srcRGB( pData[ dwPos + 0 ], pData[ dwPos + 1 ], pData[ dwPos + 2 ] );

			// equivalent in cube position 000
			RGB< dword > cRGB( dword( srcRGB.r ) * ( m_dwSize - 1 ) / 256, dword( srcRGB.g ) * ( m_dwSize - 1 ) / 256, dword( srcRGB.b ) * ( m_dwSize - 1 ) / 256 );
			Assert( cRGB.r < m_dwSize && cRGB.g < m_dwSize && cRGB.b < m_dwSize );

			// set some local small cube
			CubeLut c( 2 );
			for( dword ib = 0; ib < 2; ib++ )
				for( dword ig = 0; ig < 2; ig++ )
					for( dword ir = 0; ir < 2; ir++ )
						c._Set( ir, ig, ib, _Get( cRGB.r + ir, cRGB.g + ig, cRGB.b + ib ) );
			
			// steps for interpolation
			RGB< double > intRGB(	( double( srcRGB.r ) * double( m_dwSize - 1 ) / double( 256 ) ) - double( cRGB.r ),
									( double( srcRGB.g ) * double( m_dwSize - 1 ) / double( 256 ) ) - double( cRGB.g ),
									( double( srcRGB.b ) * double( m_dwSize - 1 ) / double( 256 ) ) - double( cRGB.b ) );
			Assert( intRGB.r >= 0 && intRGB.r <= 1 && intRGB.g >= 0 && intRGB.g <= 1 && intRGB.b >= 0 && intRGB.b <= 1 );

			// interpolate on the 3 axis
			RGB< double > rgb = _Interpolate(	_Interpolate(	_Interpolate( c._Get( 0, 0, 0 ), c._Get( 1, 0, 0 ), intRGB.r ),
																_Interpolate( c._Get( 0, 1, 0 ), c._Get( 1, 1, 0 ), intRGB.r ),
												intRGB.g ),
												_Interpolate(	_Interpolate( c._Get( 0, 0, 1 ), c._Get( 1, 0, 1 ), intRGB.r ),
																_Interpolate( c._Get( 0, 1, 1 ), c._Get( 1, 1, 1 ), intRGB.r ),
												intRGB.g ),
									intRGB.b );
			Assert( rgb.r >= 0 && rgb.r <= 1 && rgb.g >= 0 && rgb.g <= 1 && rgb.b >= 0 && rgb.b <= 1 );
			
			// set altered rgb components
			pData[ dwPos + 0 ] = byte( rgb.r * 255 );
			pData[ dwPos + 1 ] = byte( rgb.g * 255 );
			pData[ dwPos + 2 ] = byte( rgb.b * 255 );
		}
	}
}


// generate a lookup table
const bool CubeLut::Generate( const Image & _srcImg, const Image & _dstImg, const wstring & _sFileName, const dword _dwSize ) {
	Assert( _dwSize != 0 );
	Assert( _srcImg.GetWidth() != 0 && _srcImg.GetHeight() != 0 );
	
	// images must have the same dimensions
	if( _srcImg.GetWidth() != _dstImg.GetWidth() || _srcImg.GetHeight() != _dstImg.GetHeight() ) {
		wcout << L"[ERROR] source & destination images must have the same dimensions\n";
		return false;
	}

	// get images data access
	byte * pSrcData = _srcImg;
	byte * pDstData = _dstImg;

	// separate the 3 RGB channels workers
	pair< dword, dword > rgb[ 3 ][ 256 ];
	for( int i = 0; i < 3; i++ ) {
		for( int j = 0; j < 256; j++ ) {
			pair< dword, dword > & p = rgb[ i ][ j ];
			p.first = 0; p.second = 0;
		}
	}

	// for each components
	for( dword y = 0; y < _srcImg.GetHeight(); y++ ) {
		dword dwYS = y * _srcImg.GetWidth();
		for( dword x = 0; x < _srcImg.GetWidth(); x++ ) {
			dword dwPos = ( dwYS + x ) * 3;
			// get src rgb components
			RGB< byte > srcRGB( pSrcData[ dwPos + 0 ], pSrcData[ dwPos + 1 ], pSrcData[ dwPos + 2 ] );

			// get dst rgb components
			RGB< byte > dstRGB( pDstData[ dwPos + 0 ], pDstData[ dwPos + 1 ], pDstData[ dwPos + 2 ] );

			// store all displacement values with accumulators
			pair< dword, dword > & r = rgb[ 0 ][ srcRGB.r ];
			pair< dword, dword > & g = rgb[ 1 ][ srcRGB.g ];
			pair< dword, dword > & b = rgb[ 2 ][ srcRGB.b ];
			r.first += dstRGB.r; r.second++;
			g.first += dstRGB.g; g.second++;
			b.first += dstRGB.b; b.second++;
		}
	}

	// we need to add extremities values if not set (0->255) on the 3 channels
	for( int i = 0; i < 3; i++ ) {
		for( int j = 0; j < 2; j++ ) {
			int k = j * 255;
			pair< dword, dword > & p = rgb[ i ][ k ];
			if( p.second == 0 ) { p.first = k; p.second = 1; }
		}
	}

	// interpolate all values in full displacement map
	dword dwA( 0 );
	dword k = 0;
	dword dwB( 0 );
	dword l = 0;
	bool bSearch = false;
	for( dword i = 0; i < 3; i++ ) {
		for( dword j = 0; j < 256; j++ ) {
			pair< dword, dword > & p = rgb[ i ][ j ];
			if( p.second != 0 ) {
				// use the accumulator
				p.first /= p.second;
				Assert( p.first <= 255 );
				dwA = p.first;
				k = j;
				bSearch = true;
				continue;
			}

			// get the next valid value
			if( bSearch ) {
				l = j;
				while( l < 256 && bSearch ) {
					pair< dword, dword > & q = rgb[ i ][ l++ ];
					if( q.second == 0 )
						continue;
					dwB = q.first / q.second;
					bSearch = false;
					l--;
				}
			}

			// interpolate
			p.first = dword( ( double( j - k ) * ( double( dwB ) - double( dwA ) ) / ( double( l - k ) ) ) + double( dwA ) );
			Assert( p.first <= 255 );
		}
	}

	// create data stream
	ostringstream cubeDataStr;
	string sName( _sFileName.begin(), _sFileName.end() );
	size_t dotPos = sName.rfind( '.' );
	sName.erase( dotPos, sName.length() - dotPos );
	cubeDataStr << "# created with Cube tool (c)2013\n"
				<< "TITLE \"" << sName << "\"\n\n"
				<< "# " << _dwSize << " x " << _dwSize << " x " << _dwSize << " 3D lookup table\n"
				<< "LUT_3D_SIZE " << _dwSize << "\n\n"
				<< "# data domain\n"
				<< "DOMAIN_MIN 0.0 0.0 0.0\n"
				<< "DOMAIN_MAX 1.0 1.0 1.0\n\n"
				<< "# LUT data points\n";

	// reduce values to _dwSize and store
	RGB< double > col;
	for( dword ib = 0; ib < _dwSize; ib++ ) {
		col.b = double( rgb[ 2 ][ ib * 255 / ( _dwSize - 1 ) ].first ) / double( 255 );
		Assert( col.b >= 0 && col.b <= 1 );
		for( dword ig = 0; ig < _dwSize; ig++ ) {
			col.g = double( rgb[ 1 ][ ig * 255 / ( _dwSize - 1 ) ].first ) / double( 255 );
			Assert( col.g >= 0 && col.g <= 1 );
			for( dword ir = 0; ir < _dwSize; ir++ ) {
				col.r = double( rgb[ 0 ][ ir * 255 / ( _dwSize - 1 ) ].first ) / double( 255 );
				Assert( col.r >= 0 && col.r <= 1 );
				cubeDataStr << col.r << " " << col.g << " " << col.b << "\n";
			}
		}
	}

	// save file content
	ofstream dstFile;
	dstFile.open( _sFileName.c_str(), ios::out | ios::binary );
	if( !dstFile ) {
		wcout << L"[ERROR] fail to create destination file: '" << _sFileName << L"'\n";
		return false;
	}
	string cubeData( cubeDataStr.str() );
	dstFile.write( &cubeData[ 0 ], cubeData.size() );
	dstFile.close();

	// save file
	return true;
}